#!/usr/bin/python3

import numpy as np
import pandas as pd
from .. import config
import scipy.sparse as sp


def get_distance_matrix(full_att_series: pd.Series, from_ids: pd.Series, id_to_index: dict, to_ids: pd.Series = None,
                        coefficient='jaccard') -> sp.coo_matrix:
    """
    Calculating the distance of each element in from_ids to elements in to_ids, if provided, or each
    element in from_ids based on coefficient.

    :param full_att_series: dataframe with 2 columns (id, attribute values)
    :param from_ids: dataframe with 2 columns (id, attribute values)
    :param id_to_index: dict with mapping of id to index inside sparse matrix
    :param to_ids: dataframe with 2 columns (id, attribute values)
    :param coefficient: coefficient type for the distance. Possible: jaccard or overlap [Default="jaccard"]
    :return: distance sparse matrix
    """

    def get_distance(index1: int, index2: int):
        if coefficient == "jaccard":
            return jaccard_coefficient(tar_att_set=full_att_series[index1], ref_att_set=full_att_series[index2])
        else:  # coefficient == "overlap"
            return overlap_coefficient(tar_att_set=full_att_series[index1], ref_att_set=full_att_series[index2])

    row, col, data = list(), list(), list()
    # ===== from_ids against from_ids =====
    if to_ids is None:
        from_ids = from_ids.to_numpy()
        for id1_index in range(0, len(from_ids) - 1):
            for id2_index in range(id1_index + 1, len(from_ids)):
                calc_dis = get_distance(index1=id_to_index[from_ids[id1_index]],
                                        index2=id_to_index[from_ids[id2_index]])
                # assign to matrix
                if calc_dis > 0.0:
                    if id_to_index[from_ids[id1_index]] < id_to_index[from_ids[id2_index]]:
                        row.append(id_to_index[from_ids[id1_index]])
                        col.append(id_to_index[from_ids[id2_index]])
                    else:
                        row.append(id_to_index[from_ids[id2_index]])
                        col.append(id_to_index[from_ids[id1_index]])
                    data.append(calc_dis)

    # ===== from_ids against to_ids =====
    else:
        for id1 in from_ids:
            for id2 in to_ids:
                calc_dis = get_distance(index1=id_to_index[id1], index2=id_to_index[id2])
                # assign to matrix
                if calc_dis > 0.0:
                    if id_to_index[id1] < id_to_index[id2]:
                        row.append(id_to_index[id1])
                        col.append(id_to_index[id2])
                    else:
                        row.append(id_to_index[id2])
                        col.append(id_to_index[id1])
                    data.append(calc_dis)
    return sp.coo_matrix((np.array(data), (np.array(row), np.array(col))),
                         shape=(len(full_att_series), len(full_att_series)))


def create_ref_dict(mapping: pd.DataFrame, keys: set, enriched: bool = False):
    """
    Create reference dictionary with each attribute type as key
    and the union of all attribute values in the set.

    :param mapping: mapping of reference to attributes
    :param keys: attribute names
    :param enriched: bool if set resulted from enrichment analysis or not
    :return: reference dictionary with unified values
    """
    reference_dict = dict()
    if enriched:
        for key in keys:
            if key in mapping:
                reference_dict[config.ENRICH_KEY[key]] = set(mapping[key].dropna())
            else:
                reference_dict[config.ENRICH_KEY[key]] = set()
    else:
        for att_type in keys:
            reference_dict[att_type] = set.union(*mapping[att_type])
    return reference_dict


def overlap_coefficient(tar_att_set: set, ref_att_set: set):
    """
    Calculate overlap coefficient by dividing the length of overlapping elements
    of two sets by the minimum length of the two sets.

    :param tar_att_set: target set of attribute values
    :param ref_att_set: reference set of attribute values
    :return: overlap coefficient
    """
    if len(tar_att_set) == 0 & len(ref_att_set) == 0:
        return 0.0
    intersection = len(tar_att_set.intersection(ref_att_set))
    if intersection == 0:
        return 0.0
    return intersection / min(len(tar_att_set), len(ref_att_set))


def jaccard_coefficient(tar_att_set: set, ref_att_set: set):
    """
    Calculate jaccard coefficient by dividing the length of overlapping elements
    of two sets by the combined length of the two sets.

    :param tar_att_set: target set of attribute values
    :param ref_att_set: reference set of attribute values
    :return: jaccard coefficient
    """
    if len(tar_att_set) == 0 & len(ref_att_set) == 0:
        return 0.0
    intersection = len(tar_att_set.intersection(ref_att_set))
    if intersection == 0:
        return 0.0
    return intersection / len(tar_att_set.union(ref_att_set))


def calc_pvalue(test_value: dict, random_values: pd.DataFrame, maximize=True):
    """
    Calculate pvalue based on the values of the original input and the values of random runs.

    :param test_value: values from the original input as dict with attributes as key
    :param random_values: values from random runs as dataframe and attributes as columns
    :param maximize: true if the goal is to have a high test value or false if low test value [Default=True]
    :return: pvalue
    """
    pvalue = dict()
    for keys in test_value:
        pvalue[keys] = (1 + sum(test_value[keys] <= random_values[keys])) / (
                    len(random_values.index) + 1) if maximize else (1 + sum(
            test_value[keys] >= random_values[keys])) / (len(random_values.index) + 1)
    return pvalue



#!/usr/bin/python3

import pandas as pd
from collections import defaultdict


def precalc_distance_dicts(ids_cluster: pd.DataFrame, ids_mapping: pd.DataFrame, distances: dict, index_to_id: dict,
                           ids: dict):
    """
    Precalculate intra and inter distances based on pairwise distances calculated beforehand.

    :param ids_cluster: mapping of ids to corresponding cluster
    :param ids_mapping: mapping of different id types to original id type
    :param distances: all pairwise distances previously calculated
    :param index_to_id: dictionary with mapping of distance matrix indices to ids
    :param ids: dictionary with mapping of keys to mapping information in mapper
    :return: 4 dictionaries consisting of entity intra, entity inter, overall intra and overall inter distances
    """
    # ===== create empty default dicts =====
    entity_intra = defaultdict(lambda: {'max': None, 'sum': 0, 'min': None, 'count': 0})
    entity_inter = defaultdict(lambda: defaultdict(lambda: {'max': None, 'sum': 0, 'min': None, 'count': 0}))
    intra = defaultdict(lambda: {'max': None, 'sum': 0, 'min': None, 'count': 0})
    inter = defaultdict(lambda: defaultdict(lambda: {'max': None, 'sum': 0, 'min': None, 'count': 0}))
    # ===== map ids to cluster =====
    id_to_cluster = ids_cluster.set_index('id').to_dict()['cluster_index']
    # ===== map att ids to ids =====
    if ids['att_id'] != ids['id_type']:
        att_id_to_id = \
            ids_mapping[[ids['att_id'], ids['id_type']]].groupby(ids['att_id']).agg(lambda g: set(g)).to_dict()[
                ids['id_type']]
    else:
        att_id_to_id = None

    # ===== method to add values =====
    def add_value(destination, distance):
        dist = 1 - distance
        if destination['max'] is None or destination['max'] < dist:
            destination['max'] = dist
        if destination['min'] is None or destination['min'] > dist:
            destination['min'] = dist
        destination['sum'] = destination['sum'] + dist
        destination['count'] = destination['count'] + 1
        return destination

    # ===== assign distances > 0.0  to dicts =====
    for index1, index2 in distances:
        att_id1 = index_to_id[index1]
        att_id2 = index_to_id[index2]
        att_id_to_id1 = att_id_to_id[att_id1] if att_id_to_id is not None else {att_id1}
        att_id_to_id2 = att_id_to_id[att_id2] if att_id_to_id is not None else {att_id2}
        for id1 in att_id_to_id1:
            for id2 in att_id_to_id2:
                id1_cluster = id_to_cluster[id1]
                id2_cluster = id_to_cluster[id2]
                if id1_cluster == id2_cluster:
                    entity_intra[id1] = add_value(destination=entity_intra[id1], distance=distances[(index1, index2)])
                    entity_intra[id2] = add_value(destination=entity_intra[id2], distance=distances[(index1, index2)])
                    intra[id1_cluster] = add_value(destination=intra[id1_cluster], distance=distances[(index1, index2)])
                else:
                    entity_inter[id1][id2_cluster] = add_value(destination=entity_inter[id1][id2_cluster],
                                                               distance=distances[(index1, index2)])
                    entity_inter[id2][id1_cluster] = add_value(destination=entity_inter[id2][id1_cluster],
                                                               distance=distances[(index1, index2)])
                    inter[id1_cluster][id2_cluster] = add_value(destination=inter[id1_cluster][id2_cluster],
                                                                distance=distances[(index1, index2)])
                    inter[id2_cluster][id1_cluster] = add_value(destination=inter[id2_cluster][id1_cluster],
                                                                distance=distances[(index1, index2)])
    return {'entity_intra': entity_intra, 'entity_inter': entity_inter, 'intra': intra, 'inter': inter}


def calc_linkage(value_dict: dict, size: int, linkage="average"):
    """
    Calculate desired linkage based on precalculated values.

    :param value_dict: dictionary with precalculated values
    :param size: size of cluster
    :param linkage: linkage to be calculated [Default="average"]
    :return: linkage value
    """
    if linkage == "average":
        return (value_dict['sum'] + (size - value_dict['count'])) / size
    if linkage == "complete":
        return value_dict['max']
    if linkage == "single":
        return value_dict['min']
    else:
        return None


def silhouette_score(ids_cluster: pd.DataFrame, distances: dict, linkage="average"):
    """
    Calculate the silhouette score for the given cluster.

    :param ids_cluster: mapping of ids to corresponding cluster
    :param distances: all pairwise distances previously calculated
    :param linkage: linkage type for intra and inter distance [Default="average"]
    :return: sillhouette score for the whole clustering and for each cluster separately
    """
    cluster_sizes = ids_cluster['cluster_index'].value_counts().to_dict()
    # ===== map ids to cluster =====
    id_to_cluster = ids_cluster.set_index('id').to_dict()['cluster_index']
    # ===== calculate score =====
    s_score = 0
    intra_s_scores = dict()
    for entity in set(distances['entity_intra'].keys()).union(set(distances['entity_inter'].keys())):
        current_cluster = id_to_cluster[entity]
        # ===== calc intra distance =====
        if entity in distances['entity_intra']:
            entity_intra = calc_linkage(value_dict=distances['entity_intra'][entity],
                                        size=cluster_sizes[current_cluster], linkage=linkage)
        else:
            entity_intra = 1
        # ===== calc min inter distance =====
        min_entity_inter = None
        if entity in distances['entity_inter'] and len(distances['entity_inter'][entity]) < (len(cluster_sizes) - 1):
            for cluster in distances['entity_inter'][entity]:
                distance = calc_linkage(value_dict=distances['entity_inter'][entity][cluster],
                                        size=cluster_sizes[cluster], linkage=linkage)
                if min_entity_inter is None or min_entity_inter > distance:
                    min_entity_inter = distance
        if min_entity_inter is None:
            min_entity_inter = 1

        if cluster_sizes[current_cluster] > 1 and max(min_entity_inter, entity_intra) > 0.0:
            score = ((min_entity_inter - entity_intra) / max(min_entity_inter, entity_intra))
        else:
            score = 0.0
        # ===== save score for every cluster separately =====
        if current_cluster not in intra_s_scores:
            intra_s_scores[current_cluster] = 0
        intra_s_scores[current_cluster] += score
        # ===== save for total score =====
        s_score += score
    for cluster in cluster_sizes:
        if cluster in intra_s_scores:
            intra_s_scores[cluster] = intra_s_scores[cluster] / cluster_sizes[cluster]
        else:
            intra_s_scores[cluster] = 0
    return s_score / len(ids_cluster['id']), intra_s_scores


def dunn_index(ids_cluster: pd.DataFrame, distances: dict, linkage="average") -> float:
    """
    Calculate the dunn index for the given cluster.

    :param ids_cluster: mapping of ids to corresponding cluster
    :param distances: all pairwise distances previously calculated
    :param linkage: linkage type for intra and inter distance [Default="average"]
    :return: dunn index score as float
    """
    max_intra_dist = 0
    min_inter_dist = None

    # ===== count cluster size =====
    cluster_to_size = ids_cluster['cluster_index'].value_counts().to_dict()
    for cluster in cluster_to_size:
        # ===== calc intra distance =====
        if cluster in distances['intra']:
            distance = calc_linkage(value_dict=distances['intra'][cluster],
                                    size=(cluster_to_size[cluster] * cluster_to_size[cluster]) / 2, linkage=linkage)
        else:  # if no distances saved for intra cluster: all pairwise have distance 1
            distance = 1
        if max_intra_dist is None or max_intra_dist < distance:
            max_intra_dist = distance
        # ===== calc min inter distance =====
        if cluster in distances['inter']:
            # ===== distance to at least one cluster == 1 -> not in dict =====
            if len(distances['inter'][cluster]) < (len(cluster_to_size) - 1):
                min_inter_dist = 1
            # ===== calc distance to all clusters =====
            for to_cluster in distances['inter'][cluster]:
                distance = calc_linkage(value_dict=distances['inter'][cluster][to_cluster],
                                        size=cluster_to_size[cluster] * cluster_to_size[to_cluster], linkage=linkage)
                if min_inter_dist is None or min_inter_dist > distance:
                    min_inter_dist = distance
    if max_intra_dist == 0 or min_inter_dist is None:
        return 0.0
    return min_inter_dist / max_intra_dist


def davies_bouldin_index(ids_cluster: pd.DataFrame, distances: dict, linkage="average") -> float:
    """
    Calculate the davies bouldin index for the given cluster.

    :param ids_cluster: mapping of ids to corresponding cluster
    :param distances: all pairwise distances previously calculated
    :param linkage: linkage type for intra and inter distance [Default="average"]
    :return: davies bouldin index score as float
    """
    max_values = 0
    # ===== count cluster size =====
    cluster_to_size = ids_cluster['cluster_index'].value_counts().to_dict()
    for cluster_i in cluster_to_size:
        cur_max = 0
        for cluster_j in cluster_to_size:
            if cluster_i != cluster_j:
                # ===== calc intra distances =====
                distance_i = calc_linkage(value_dict=distances['intra'][cluster_i],
                                          size=(cluster_to_size[cluster_i] * cluster_to_size[cluster_i]) / 2,
                                          linkage=linkage) if cluster_i in distances['intra'] else 1
                distance_j = calc_linkage(value_dict=distances['intra'][cluster_j],
                                          size=(cluster_to_size[cluster_j] * cluster_to_size[cluster_j]) / 2,
                                          linkage=linkage) if cluster_j in distances['intra'] else 1
                # ===== calc inter distance =====
                distance_ij = calc_linkage(value_dict=distances['inter'][cluster_i][cluster_j],
                                           size=cluster_to_size[cluster_i] * cluster_to_size[cluster_j],
                                           linkage=linkage)
                # ===== calc (q(ci)*q(cj))/p(ci,cj) =====
                value = (distance_i * distance_j) / distance_ij if distance_ij != 0 else 0
                if value > cur_max:
                    cur_max = value
        max_values = max_values + cur_max
    return max_values / len(cluster_to_size)



#!/usr/bin/python3

import os
import math
import numpy as np
import pandas as pd
from .. import config as c
import seaborn as sns
from pathlib import Path
from matplotlib import pyplot as plt
from collections import defaultdict
from ..mappers.mapper import Mapper, FileMapper

sns.set_palette("colorblind")

plt.rcParams.update({'font.size': 17, 'axes.titlelocation': "left", 'axes.titleweight': "bold", 'axes.labelsize': 21})
plt.rcParams['figure.dpi'] = 300
plt.rcParams['savefig.dpi'] = 300

replacements = {"related_genes": "related\ngenes", "related_variants": "related\nvariants"}
eval_terms = {"DI-based": {"single": "Dunn index", "multi": "Dunn indices"},
              "SS-based": {"single": "Sillhouette score", "multi": "Sillhouette scores"},
              "DBI-based": {"single": "Davis-Bouldin index", "multi": "Davis-Bouldin indices"},
              "JI-based": {"single": "Jaccard index", "multi": "Jaccard indices"},
              "OC-based": {"single": "Overlap coefficient", "multi": "Overlap coefficients"}}
annot_terms = {'GO.BP': 'GO.BP-based', 'GO.CC':'GO.CC-based', 'GO.MF':'GO.MF-based',
               'KEGG': 'KEGG-based', "related_genes": "associated genes", "related_variants": "associated variants"}

def create_plots(results, mode, tar, tar_id, out_dir, prefix, file_type: str = "pdf"):
    """

    :param results: results generated from single_validation method
    :param mode: comparison mode [set, id-set, set-set, cluster]
    :param tar: path to the file with the target input
    :param tar_id: id type of target input
    :param out_dir: output directory for results
    :param prefix: prefix for the file name
    :param file_type: file ending the plots should have [Default=pdf]
    :return:
    """
    Path(out_dir).mkdir(parents=True, exist_ok=True)  # make sure output dir exists
    if mode == "clustering":
        cluster_plot(results=results, user_input={"clustering": tar, "type": tar_id},
                     out_dir=out_dir, prefix=prefix, file_type=file_type)
    else:
        set_plot(results=results, user_input={"set": tar, "type": tar_id},
                 out_dir=out_dir, prefix=prefix, file_type=file_type)


def cluster_plot(results, user_input, out_dir, prefix, file_type: str = "pdf"):
    in_type = "diseases" if user_input["type"] in c.SUPPORTED_DISEASE_IDS else "genes"
    # ===== Prepare for scatterplot =====
    p_value_df = pd.DataFrame.from_dict(results["p_values"]["values"]).rename_axis('attribute').reset_index()
    p_value_df = p_value_df.replace(replacements).sort_values(['attribute']).reset_index(drop=True)
    for val in results["p_values"]["values"]:
        p_value_df["log_p-values"] = p_value_df[val].apply(lambda x: -math.log10(x))
        # ===== Plot scatterplot =====
        p_value_plot(title="Empirical P-value\nbased on " +eval_terms[val]["single"], p_value_df=p_value_df, out_dir=out_dir,
                     prefix=prefix + "_" + val, file_type=file_type)
    # ===== Prepare for mappability plot =====
    mapped_df = user_input["clustering"][['id', 'cluster']]
    cluster_sizes = mapped_df['cluster'].value_counts().to_dict()
    for att in results["input_values"]["mapped_ids"]:
        mapped_df[att] = [1 if x in results["input_values"]["mapped_ids"][att] else 0 for x in mapped_df['id']]
    mapped_df = mapped_df.groupby('cluster', as_index=False).agg(sum).melt('cluster', var_name='attribute',
                                                                           value_name='count')
    mapped_df = mapped_df.replace(replacements).sort_values(['attribute']).reset_index(drop=True)
    mapped_df["fraction"] = mapped_df.apply(lambda x: x['count'] / cluster_sizes[x['cluster']], axis=1)
    # ===== Plot mappability plot =====
    mappability_plot(title="Mappability of input\ninput to annotations", in_type=in_type, mapped_df=mapped_df, out_dir=out_dir,
                     prefix=prefix, cluster=True, file_type=file_type)


def set_plot(results, user_input, out_dir, prefix, file_type: str = "pdf"):
    in_type = "diseases" if user_input["type"] in c.SUPPORTED_DISEASE_IDS else "genes"
    # ===== Prepare for scatterplot =====
    p_value_df = pd.DataFrame.from_dict(
        {'p_values': results["p_values"]["values"][next(iter(results["p_values"]["values"]))]}).rename_axis(
        'attribute').reset_index()
    p_value_df["log_p-values"] = p_value_df["p_values"].apply(lambda x: -math.log10(x))
    p_value_df = p_value_df.replace(replacements).sort_values(['attribute']).reset_index(drop=True)
    # ===== Plot scatterplot =====
    for val in results["p_values"]["values"]:
        p_value_plot(title="Empirical P-value\nbased on " +eval_terms[val]["single"], p_value_df=p_value_df,
                     out_dir=out_dir, prefix=prefix + "_" + val, file_type=file_type)
    # ===== Prepare for mappability plot =====
    mapped_df = pd.DataFrame()
    for att in results["input_values"]["mapped_ids"]:
        mapped_df[att] = [1 if x in results["input_values"]["mapped_ids"][att] else 0 for x in user_input["set"]]
    mapped_df = mapped_df.T
    mapped_df["count"] = mapped_df.sum(axis=1)
    mapped_df["fraction"] = mapped_df['count'].apply(lambda x: x / len(user_input["set"]))
    mapped_df = mapped_df.rename_axis('attribute').reset_index()
    mapped_df = mapped_df.replace(replacements).sort_values(['attribute']).reset_index(drop=True)
    # ===== Plot mappability plot =====
    mappability_plot(title="Mappability of\ninput to annotations", in_type=in_type, mapped_df=mapped_df, out_dir=out_dir,
                     prefix=prefix, cluster=False, file_type=file_type)


def p_value_plot(title, p_value_df, out_dir, prefix, file_type: str = "pdf"):
    fig = plt.figure(figsize=(6, 6), dpi=80)
    ax = sns.scatterplot(x=p_value_df['attribute'], y=p_value_df['log_p-values'], s=150)
    ax.set(title=title, ylabel="-log10(P)", xlabel="", ylim=(0, 3.1))
    ax.axhline(y=-math.log10(0.05), color="red", linestyle='--')
    fig.tight_layout()
    fig.savefig(os.path.join(out_dir, prefix + '_p-value.' + file_type), bbox_inches='tight')


def mappability_plot(title, in_type, mapped_df, out_dir, prefix, cluster=False, file_type: str = "pdf"):
    if cluster:
        fig = plt.figure(figsize=(7, 6), dpi=80)
        ax = sns.barplot(x="attribute", y='fraction', data=mapped_df, hue="cluster")
    else:
        fig = plt.figure(figsize=(6, 6), dpi=80)
        ax = sns.barplot(x="attribute", y='fraction', data=mapped_df)
    ax.set(title=title, xlabel="", ylabel="Fraction of " + in_type + " with\nnon-empty annotation sets", ylim=(0, 1.1))
    if cluster:
        ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), title="Cluster")
    fig.tight_layout()
    fig.savefig(os.path.join(out_dir, prefix + '_mappability.' + file_type), bbox_inches='tight')


def create_extended_plots(results, mode, tar, out_dir, prefix, file_type: str = "pdf", mapper:Mapper = FileMapper()):
    """

    :param results: results generated from single_validation method
    :param mode: comparison mode [set, id-set, set-set, cluster]
    :param tar: path to the file with the target input
    :param tar_id: id type of target input
    :param out_dir: output directory for results
    :param prefix: prefix for the file name
    :param file_type: file ending the plots should have [Default=pdf]
    :return:
    """
    Path(out_dir).mkdir(parents=True, exist_ok=True)  # make sure output dir exists
    value_distribution_plots(results=results, out_dir=out_dir, prefix=prefix, file_type=file_type)
    term_annotation_plots(results=results, out_dir=out_dir, prefix=prefix, file_type=file_type)
    sankey_plot(results=results, mode=mode, mapper=mapper, out_dir=out_dir, prefix=prefix,
                file_type=file_type, tar_cluster=tar)


def value_distribution_plots(results, out_dir, prefix, file_type: str = "pdf"):
    for eval_term in results["input_values"]['values']:
        df = pd.melt(pd.DataFrame(results['random_values']['values'][eval_term]))
        df['value'] = df['value'].astype(float)
        for term_index, term in enumerate(results["input_values"]['values'][eval_term]):
            fig = plt.figure(figsize=(7, 6), dpi=80)
            plt.axvline(float(results["input_values"]['values'][eval_term][term]), color='darkred', lw=10)
            ax = sns.histplot(df[df["variable"] == term], x="value", kde=True, color=sns.color_palette()[term_index],
                              bins=10)
            if term in replacements:
                plt.title("Distribution of " + eval_terms[eval_term]["multi"] + "\nbased on " + annot_terms[term])
                plt.xlabel(eval_terms[eval_term]["single"] + " based\non " + annot_terms[term])
            else:
                plt.title("Distribution of\n" + annot_terms[term] + " " + eval_terms[eval_term]["multi"])
                plt.xlabel(annot_terms[term] + " " + eval_terms[eval_term]["single"])
            plt.ylabel("Number of runs\non randomized data")
            props = dict(boxstyle='round', facecolor='wheat', alpha=0.5)
            ax.text(0.05, 0.95, "Empirical\nP-value:\n%.3f" % results["p_values"]['values'][eval_term][term],
                    transform=ax.transAxes, fontsize=14,
                    verticalalignment='top', bbox=props)
            fig.tight_layout()
            fig.savefig(os.path.join(out_dir, prefix + '_' + eval_term + '_' + term + '_distribution.' + file_type),
                        bbox_inches='tight')


def term_annotation_plots(results, out_dir, prefix, file_type: str = "pdf"):
    df = pd.DataFrame(results["input_values"]["mapped_ids"]).fillna("").applymap(len)
    for term_index, term in enumerate(df.columns):
        fig = plt.figure(figsize=(7, 6), dpi=80)
        sns.histplot(df, x=term, kde=True, color=sns.color_palette()[term_index], bins=10)
        if term in replacements:
            plt.title("Distribution of numbers\nof " + annot_terms[term])
            plt.xlabel("Number of " + annot_terms[term])
        else:
            plt.title("Distribution of numbers\nof associated " + term + "-terms")
            plt.xlabel("Number of associated\n" + term + "-terms")
        plt.ylabel("Number of IDs\ncontained in query")
        fig.tight_layout()
        fig.savefig(os.path.join(out_dir, prefix + '_' + term + '_annotation_distribution.' + file_type),
                    bbox_inches='tight')


def sankey_plot(results, mode, out_dir, prefix, file_type: str = "pdf", tar_cluster=None, include_others=False,
                mapper:Mapper = FileMapper()):
    full_df = pd.DataFrame(results["input_values"]["mapped_ids"])
    for term_index, term in enumerate(full_df.columns):

        def col_sort(df, colname, hierarchy):
            df[colname] = df[colname].astype("category")
            df[colname].cat.set_categories(hierarchy, inplace=True)
            df = df.sort_values([colname], ascending=True)
            return df

        d = full_df[[term]].dropna().rename_axis('left').reset_index().explode(term)
        if mode == "clustering":
            d = d.replace({"left": tar_cluster.set_index('id')["cluster"].to_dict()})
            df = pd.DataFrame(columns=["id", term])
            for cluster in d['left'].unique():
                df = pd.concat([df, d[d["left"] == cluster][term].value_counts(normalize=True).rename_axis(
                    'id').reset_index(name=term)])
            ids = df.groupby(['id']).sum().rename_axis('id').reset_index().nlargest(10, term)["id"]
        else:
            ids = d[term].value_counts().nlargest(10).index
        if include_others:
            d.loc[~d[term].isin(ids), term] = "other"
        else:
            d = d[d[term].isin(ids)]
        d[term] = d[term].astype({term: str}, errors='raise')
        if term == "related_genes":
            rename_ids, _ = mapper.get_loaded_mapping(in_set=d[term], id_type="entrezgene", key="gene_ids")
            rename_ids = rename_ids.explode("symbol").set_index('entrezgene')['symbol'].to_dict()
            d[term] = d[term].map(rename_ids).fillna(d[term])
        # ===== Save hierarchy =====
        hierarchy_left = d["left"].value_counts().index.tolist()
        hierarchy_right = d[term].value_counts().index.tolist()
        # ===== Sort =====
        d = d.reset_index(drop=True).value_counts().reset_index()
        d = col_sort(df=d, colname="left", hierarchy=hierarchy_left)
        d = col_sort(df=d, colname=term, hierarchy=hierarchy_right)
        # ===== Assign colors =====
        z = {**dict(zip(d["left"].unique(), ["#808080"] * len(d.index.unique()))),
             **dict(zip(d[term].unique(), sns.color_palette() * 10))}

        d.rename(columns={term: 'right', 0: 'weight'}, inplace=True)
        # ===== Plot =====
        sankey(data=d, aspect=20, right_color=True, color_dict=z, term=term,
               out_dir=out_dir, prefix=prefix, file_type=file_type)


def sankey(data, out_dir, prefix, file_type: str = "pdf", color_dict=None, aspect=4, right_color=False, term=None):
    """
    Make Sankey Diagram showing flow from left-->right

    :param color_dict: dictionary of colors to use for each label {'label':'color'}
    :param aspect: vertical extent of the diagram in units of horizontal extent
    :param right_color: if true, each strip in the diagram will be be colored according to its left label

    original: https://github.com/anazalea/pySankey
    """

    # ===== Identify all labels that appear 'left' or 'right' =====
    all_labels = pd.Series(np.r_[data.left.unique(), data.right.unique()]).unique()

    # ===== Identify left and right labels =====
    left_labels = data.left.unique()
    right_labels = data.right.unique()

    # ===== If no color_dict given, make one =====
    if color_dict is None:
        color_dict = {}
        palette = "hls"
        color_palette = sns.color_palette(palette, len(all_labels))
        for i, label in enumerate(all_labels):
            color_dict[label] = color_palette[i]

    # ===== Determine widths of individual strips =====
    ns_l = defaultdict()
    ns_r = defaultdict()
    for left_label in left_labels:
        left_dict = {}
        right_dict = {}
        for right_label in right_labels:
            left_dict[right_label] = data[(data.left == left_label) & (data.right == right_label)].weight.sum()
            right_dict[right_label] = data[(data.left == left_label) & (data.right == right_label)].weight.sum()
        ns_l[left_label] = left_dict
        ns_r[left_label] = right_dict

    # ===== Determine positions of left label patches and total widths =====
    left_widths = defaultdict()
    for i, left_label in enumerate(left_labels):
        myD = {}
        myD['left'] = data[data.left == left_label].weight.sum()
        if i == 0:
            myD['bottom'] = 0
            myD['top'] = myD['left']
        else:
            myD['bottom'] = left_widths[left_labels[i - 1]]['top'] + 0.02 * data.weight.sum()
            myD['top'] = myD['bottom'] + myD['left']
            top_edge = myD['top']
        left_widths[left_label] = myD

    # ===== Determine positions of right label patches and total widths =====
    right_widths = defaultdict()
    for i, right_label in enumerate(right_labels):
        myD = {}
        myD['right'] = data[data.right == right_label].weight.sum()
        if i == 0:
            myD['bottom'] = 0
            myD['top'] = myD['right']
        else:
            myD['bottom'] = right_widths[right_labels[i - 1]]['top'] + 0.02 * data.weight.sum()
            myD['top'] = myD['bottom'] + myD['right']
            top_edge = myD['top']
        right_widths[right_label] = myD

    # ===== Total vertical extent of diagram =====
    xMax = top_edge / aspect

    # ===== Draw vertical bars on left and right of each  label's section & print label =====
    fig = plt.figure(dpi=80)
    for left_label in left_labels:
        plt.fill_between(
            [-0.02 * xMax, 0],
            2 * [left_widths[left_label]['bottom']],
            2 * [left_widths[left_label]['bottom'] + left_widths[left_label]['left']],
            color=color_dict[left_label],
            alpha=0.99
        )
        plt.text(
            -0.05 * xMax,
            left_widths[left_label]['bottom'] + 0.5 * left_widths[left_label]['left'],
            left_label,
            {'ha': 'right', 'va': 'center'},
            fontsize=min(left_widths[left_label]['left'] + 9, 14)
        )
    for right_label in right_labels:
        plt.fill_between(
            [xMax, 1.02 * xMax], 2 * [right_widths[right_label]['bottom']],
                                 2 * [right_widths[right_label]['bottom'] + right_widths[right_label]['right']],
            color=color_dict[right_label],
            alpha=0.99
        )
        plt.text(
            1.05 * xMax,
            right_widths[right_label]['bottom'] + 0.5 * right_widths[right_label]['right'],
            right_label,
            {'ha': 'left', 'va': 'center'},
            fontsize=min(right_widths[right_label]['right'] + 9, 14)
        )

    # ===== Plot strips =====
    for left_label in left_labels:
        for right_label in right_labels:
            label_color = left_label
            if right_color:
                label_color = right_label
            if len(data[(data.left == left_label) & (data.right == right_label)]) > 0:
                # Create array of y values for each strip, half at left value,
                # half at right, convolve
                ys_d = np.array(50 * [left_widths[left_label]['bottom']] + 50 * [right_widths[right_label]['bottom']])
                ys_d = np.convolve(ys_d, 0.05 * np.ones(20), mode='valid')
                ys_d = np.convolve(ys_d, 0.05 * np.ones(20), mode='valid')
                ys_u = np.array(50 * [left_widths[left_label]['bottom'] + ns_l[left_label][right_label]] + 50 * [
                    right_widths[right_label]['bottom'] + ns_r[left_label][right_label]])
                ys_u = np.convolve(ys_u, 0.05 * np.ones(20), mode='valid')
                ys_u = np.convolve(ys_u, 0.05 * np.ones(20), mode='valid')

                # Update bottom edges at each label so next strip starts at the right place
                left_widths[left_label]['bottom'] += ns_l[left_label][right_label]
                right_widths[right_label]['bottom'] += ns_r[left_label][right_label]
                plt.fill_between(
                    np.linspace(0, xMax, len(ys_d)), ys_d, ys_u, alpha=0.65,
                    color=color_dict[label_color]
                )
    plt.gca().axis('off')
    plt.gcf().set_size_inches(6, 6)
    if term in replacements:
        plt.title("Top 10 most frequent\n" + annot_terms[term] + "\nlinked to each ID")
    else:
        plt.title("Top 10 most frequent\n" + term + "-terms\nlinked to each ID")
    fig.tight_layout()
    fig.savefig(os.path.join(out_dir, prefix + '_' + term + '_sankey.' + file_type),
                bbox_inches='tight')



#!/usr/bin/python3

from .d_utils import eval_utils as eu
from . import config as c
from .mappers import gene_getter as gg, disease_getter as dg
from .mappers.mapper import Mapper
from . import score_calculator as sc
from abc import abstractmethod


class Comparator:
    def __init__(self, mapper: Mapper, distance_measure: str, verbose: bool = False):
        self.mapper = mapper
        self.verbose = verbose
        self.distance_measure = distance_measure
        self.input_run = True
        self.mapping = None
        self.id_set, self.id_type = None, None
        self.att_id, self.att_key, self.sparse_key = None, None, None

    def load_target(self, id_set, id_type):
        self.id_set = id_set
        self.id_type = id_type
        if id_type in c.SUPPORTED_DISEASE_IDS:
            self.mapping = dg.get_disease_to_attributes(disease_set=id_set, id_type=id_type, mapper=self.mapper)
            self.sparse_key, self.att_key, self.att_id = 'disease_mat_ids', 'disorder_atts', 'mondo'
        else:  # if id_type in c.SUPPORTED_GENE_IDS:
            self.mapping = gg.get_gene_to_attributes(gene_set=id_set, id_type=id_type, mapper=self.mapper)
            self.sparse_key, self.att_key, self.att_id = 'gene_mat_ids', 'gene_atts', 'entrezgene'

    @abstractmethod
    def compare(self, threshold: float = 0.0):
        pass


class SetComparator(Comparator):
    """
    Compare the set on itself based on connected attributes. See config for more info.
    """

    def compare(self, threshold: float = 0.0):
        result, mapped = dict(), dict()
        new_ids = self.mapper.update_distance_ids(in_series=self.mapper.loaded_mappings[self.att_key][self.att_id],
                                                  key=self.sparse_key, distance_measure=self.distance_measure)
        for attribute in self.mapping.columns[1:]:
            subset_df = self.mapping[self.mapping[attribute].str.len() > 0]
            missing_values = len(self.mapping) - len(subset_df)
            if missing_values > 0:
                print("Missing values for " + attribute + " :" + str(missing_values) + "/" + str(
                    len(self.id_set))) if self.verbose else None
            if len(new_ids) > 0:
                comp_mat = eu.get_distance_matrix(full_att_series=self.mapper.loaded_mappings[self.att_key][attribute],
                                                  from_ids=self.mapper.loaded_mappings[self.att_key][self.att_id],
                                                  id_to_index=self.mapper.loaded_distance_ids[self.distance_measure][self.sparse_key],
                                                  to_ids=new_ids)
                self.mapper.update_distances(in_mat=comp_mat, key=c.DISTANCES[attribute], id_type=self.sparse_key,
                                             distance_measure=self.distance_measure)
            if subset_df.empty:
                result[c.replacements[attribute]] = 0

            else:
                ids = self.mapper.get_loaded_mapping_ids(in_ids=set(subset_df[subset_df.columns[0]]),
                                                         id_type=self.id_type)

                if self.att_id != c.ID_TYPE_KEY[self.id_type]:
                    ids = ids[[self.att_id, c.ID_TYPE_KEY[self.id_type]]].drop_duplicates()
                sub_mat = self.mapper.get_loaded_distances(in_series=ids[self.att_id],  id_type=self.sparse_key,
                                                           key=c.DISTANCES[attribute],
                                                           distance_measure=self.distance_measure)
                axis = (len(self.mapping)-len(ids[c.ID_TYPE_KEY[self.id_type]].unique())) + len(ids)
                #missing_distances = ((axis * (axis-1) ) / 2) - sub_mat.getnnz()
                #result[c.replacements[attribute]] = ((sub_mat.getnnz() - sub_mat.sum()) + missing_distances) / \
                #                                      ((axis * (axis - 1)) / 2)
                if sub_mat.sum() == 0 or axis <= 1:
                    result[c.replacements[attribute]] = 0.0
                else:
                    result[c.replacements[attribute]] = sub_mat.sum() / ((axis * (axis - 1)) / 2)
                if self.input_run:
                    save_mapping = subset_df.copy()
                    save_mapping[attribute] = save_mapping[attribute].apply(lambda x: list(x))
                    mapped[c.replacements[attribute]] = save_mapping.set_index(c.ID_TYPE_KEY[self.id_type])[
                        attribute].to_dict()
        return result, mapped


class SetSetComparator(Comparator):
    """
    Compare two sets of the same type (either genes or diseases) with each other.
    The tar set is evaluated how good it matches to the ref set.
    """

    def __init__(self, mapper: Mapper, distance_measure: str, enriched: bool = False, verbose: bool = False):
        super().__init__(mapper=mapper, verbose=verbose, distance_measure=distance_measure)
        self.enriched = enriched
        self.ref_dict = dict()

    def load_reference(self, ref, ref_id_type, tar_id_type):
        if ref_id_type in c.SUPPORTED_DISEASE_IDS:
            id_mapping = dg.get_disease_to_attributes(disease_set=ref, id_type=ref_id_type, mapper=self.mapper)
            if not id_mapping.empty:  # Only if id mapping is not empty
                if tar_id_type in c.SUPPORTED_DISEASE_IDS:
                    self.ref_dict = eu.create_ref_dict(mapping=id_mapping, keys=id_mapping.columns[1:])
                else:  # if targets_id_type in c.SUPPORTED_GENE_IDS:
                    id_mapping = id_mapping.rename(columns={'ctd.pathway_related_to_disease': 'pathway.kegg'})
                    self.ref_dict = eu.create_ref_dict(mapping=id_mapping, keys={'pathway.kegg'})
        else:  # if targets_id_type in c.SUPPORTED_GENE_IDS:
            if self.enriched:
                id_mapping = gg.get_enriched_attributes(gene_set=ref, id_type=ref_id_type, mapper=self.mapper)
            else:
                id_mapping = gg.get_gene_to_attributes(gene_set=ref, id_type=ref_id_type, mapper=self.mapper)
            if not id_mapping.empty:  # Only if id mapping is not empty
                if tar_id_type in c.SUPPORTED_DISEASE_IDS:
                    col_name = 'KEGG_2016' if self.enriched else 'pathway.kegg'
                    id_mapping = id_mapping.rename(columns={col_name: 'ctd.pathway_related_to_disease'})
                    self.ref_dict = eu.create_ref_dict(mapping=id_mapping, keys={'ctd.pathway_related_to_disease'})
                else:  # if targets_id_type in c.SUPPORTED_GENE_IDS:
                    if self.enriched:
                        self.ref_dict = eu.create_ref_dict(mapping=id_mapping, keys=c.ENRICH_KEY.keys(), enriched=True)
                    else:
                        self.ref_dict = eu.create_ref_dict(mapping=id_mapping, keys=c.ENRICH_KEY.values(), enriched=False)

    def compare(self, threshold: float = 0.0):
        evaluation, mapped = dict(), dict()
        for attribute in self.ref_dict.keys():
            if self.distance_measure == "jaccard":
                evaluated_series = self.mapping[attribute].apply(eu.jaccard_coefficient,
                                                                 ref_att_set=self.ref_dict[attribute])
            else:  # == "overlap_coefficient"
                evaluated_series = self.mapping[attribute].apply(eu.overlap_coefficient,
                                                                 ref_att_set=self.ref_dict[attribute])
            evaluation[c.replacements[attribute]] = str(len(evaluated_series[evaluated_series > threshold]) /
                                                        len(evaluated_series))
            if self.input_run:
                save_mapping = self.mapping[self.mapping[attribute] != ""].copy()
                save_mapping[attribute] = save_mapping[attribute].apply(lambda x: list(x))
                mapped[c.replacements[attribute]] = save_mapping.set_index(c.ID_TYPE_KEY[self.id_type])[
                    attribute].to_dict()
        return evaluation, mapped


class ClusterComparator(Comparator):
    """
    Evaluate the quality of clustering of given set with diseases or genes and
    assigned clusters. Additionally calculate statistical values with
    silhouette score and dunn index.
    """

    def __init__(self, mapper: Mapper, distance_measure: str, verbose: bool = False):
        super().__init__(mapper=mapper, verbose=verbose, distance_measure=distance_measure)
        self.clustering = None

    def load_target(self, id_set, id_type):
        id_set['cluster_index'] = id_set.groupby('cluster').ngroup()
        super().load_target(id_set=id_set["id"], id_type=id_type)
        self.clustering = id_set[['id', 'cluster', 'cluster_index']]

    def compare(self, threshold: float = 0.0):
        result_di, result_ss, result_ss_intermediate, result_dbi, mapped = dict(), dict(), dict(), dict(), dict()
        new_ids = self.mapper.update_distance_ids(in_series=self.mapper.loaded_mappings[self.att_key][self.att_id],
                                                  key=self.sparse_key, distance_measure=self.distance_measure)
        for attribute in self.mapping.columns[1:]:
            subset_df = self.mapping[self.mapping[attribute].str.len() > 0]
            subset_clusters = self.clustering[self.clustering['id'].isin(subset_df[c.ID_TYPE_KEY[self.id_type]])][
                ['id', 'cluster_index']]
            missing_values = len(self.mapping) - len(subset_df)
            if missing_values > 0:
                print("Missing values for " + attribute + " :" + str(missing_values) + "/" + str(
                    len(self.mapping)) + "") if self.verbose else None

            if len(new_ids) > 0:
                comp_mat = eu.get_distance_matrix(full_att_series=self.mapper.loaded_mappings[self.att_key][attribute],
                                                  from_ids=self.mapper.loaded_mappings[self.att_key][self.att_id],
                                                  id_to_index=self.mapper.loaded_distance_ids[self.distance_measure][
                                                      self.sparse_key],
                                                  to_ids=new_ids)
                self.mapper.update_distances(in_mat=comp_mat, key=c.DISTANCES[attribute], id_type=self.sparse_key,
                                             distance_measure=self.distance_measure)

            if subset_df.empty:
                result_di[c.replacements[attribute]], result_ss[c.replacements[attribute]] = None, None
                result_ss_intermediate[c.replacements[attribute]] = None
                result_dbi[c.replacements[attribute]], mapped[c.replacements[attribute]] = None, []
            else:
                ids = self.mapper.get_loaded_mapping_ids(in_ids=set(subset_df[subset_df.columns[0]]),
                                                         id_type=self.id_type)
                distances = self.mapper.get_loaded_distances(in_series=ids[self.att_id].drop_duplicates(),
                                                             id_type=self.sparse_key,
                                                             key=c.DISTANCES[attribute],
                                                             distance_measure=self.distance_measure)
                distances = dict(distances.todok().items())
                inv_index_to_id = {index: value for index, value in ids[self.att_id].reset_index(drop=True).items()}
                precalc_dist = sc.precalc_distance_dicts(ids_cluster=subset_clusters, ids_mapping=ids,
                                                         distances=distances,
                                                         index_to_id=inv_index_to_id,
                                                         ids={'id_type': c.ID_TYPE_KEY[self.id_type],
                                                              'sparse_key': self.sparse_key, 'att_id': self.att_id,
                                                              'attribute': c.DISTANCES[attribute]})
                ss_score = sc.silhouette_score(ids_cluster=subset_clusters, distances=precalc_dist, linkage="average")
                di_score = sc.dunn_index(ids_cluster=subset_clusters, distances=precalc_dist, linkage="average")
                dbi_score = sc.davies_bouldin_index(ids_cluster=subset_clusters, distances=precalc_dist,
                                                    linkage="average")
                result_di[c.replacements[attribute]] = di_score
                result_ss[c.replacements[attribute]] = ss_score[0]
                result_ss_intermediate[c.replacements[attribute]] = ss_score[1]
                result_dbi[c.replacements[attribute]] = dbi_score
                if self.input_run:
                    save_mapping = subset_df.copy()
                    save_mapping[attribute] = save_mapping[attribute].apply(lambda x: list(x))
                    mapped[c.replacements[attribute]] = save_mapping.set_index(c.ID_TYPE_KEY[self.id_type])[
                        attribute].to_dict()
        return result_di, result_ss, result_dbi, result_ss_intermediate, mapped


#!/bin/python3

import os
import sys
import argparse
import time
import psutil
from .. import config as c
import pandas as pd

start_time = time.time()


def save_parameters(script_desc: str, arguments):
    """
    Save command line options into local variables.

    :return: values assigned to input arguments
    """
    descr = "\n############################################################################\n"
    descr += "###################### DIGEST - %(prog)s ########################\n"
    descr += script_desc
    descr += "\n############################################################################\n"
    descr += "\nusage: python3 %(prog)s [required arguments] [optional arguments]\n"
    epilo = _get_epilog(script_name=os.path.basename(sys.argv[0]))
    parser = argparse.ArgumentParser(description=descr, formatter_class=argparse.RawTextHelpFormatter, epilog=epilo,
                                     usage=argparse.SUPPRESS, add_help=False)
    required_args = parser.add_argument_group("required arguments")
    if 'r' in arguments:
        required_args.add_argument('-r', '--reference', type=str, default=None,
                                   help='[Only for mode set-set] Reference file or id. ')
    if 'ri' in arguments:
        required_args.add_argument('-ri', '--reference_id_type', type=str, default=None,
                                   choices=c.SUPPORTED_GENE_IDS + c.SUPPORTED_DISEASE_IDS, metavar='REFERENCE_ID_TYPE',
                                   help='[Only for mode set-set] Reference id type. See possible options below.')
    if 't' in arguments:
        required_args.add_argument('-t', '--target', type=str, required=True,
                                   help='Target file with set or clusters.')
    if 'ti' in arguments:
        required_args.add_argument('-ti', '--target_id_type', type=str, required=True,
                                   choices=c.SUPPORTED_GENE_IDS + c.SUPPORTED_DISEASE_IDS, metavar='TARGET_ID_TYPE',
                                   help='Target id type. See possible options below.')
    if 'm' in arguments:
        required_args.add_argument('-m', '--mode', type=str, required=True,
                                   choices=['set', 'set-set', 'clustering', 'subnetwork', 'subnetwork-set'],
                                   help='Desired mode. See possible options below.')

    optional_args = parser.add_argument_group("optional arguments")
    if 'o' in arguments:
        optional_args.add_argument('-o', '--out_dir', type=str, default='./', help='Output directory. [Default=./]')
    if 'dg' in arguments:
        optional_args.add_argument('-dg', '--distance_measure', type=str, default='jaccard',
                                   choices=["jaccard", "overlap"],
                                   help="Distance measure. [Default=jaccard]")
    if 'e' in arguments:
        optional_args.add_argument("-e", "--enriched", action='store_true', default=False,
                                   help="Set flag, if only enriched attributes of the reference should be used.")
    if 'c' in arguments:
        optional_args.add_argument("-c", "--runs", type=int, default=c.NUMBER_OF_RANDOM_RUNS,
                                   help="Number of runs with random target values for p-value calculation.")
    if 'b' in arguments:
        optional_args.add_argument("-b", "--background_model", type=str, default="complete",
                                   choices=['complete', 'term-pres', 'network'],
                                   help="Model defining how random values should be picked. See possible options below.")
    if 'n' in arguments:
        optional_args.add_argument("-n", "--network", type=str, default=None,
                                   help="Network file as sif, graphml or gt.")
    if 'ni' in arguments:
        optional_args.add_argument("-ni", "--network_id_type", type=str, default=None,
                                   help="Type of node IDs inside given network.")
    if 'np' in arguments:
        optional_args.add_argument("-np", "--network_property_name", type=str, default=None,
                                   help="If network is of graphml or gt type, enter name of vertex property with IDs.")
    if 'pr' in arguments:
        optional_args.add_argument("-pr", "--replace", type=int, default=100,
                                   help="Percentage of how many of the original ids should be replaced with random ids."
                                        " [Default=100]")
    if 'v' in arguments:
        optional_args.add_argument("-v", "--verbose", action='store_true', default=False,
                                   help="Set flag, if additional info like ids without assigned attributes should "
                                        "be printed.")
    if 'p' in arguments:
        optional_args.add_argument("-p", "--plot", action='store_true', default=False,
                                   help="Set flag, if plots should be created.")
    if 's' in arguments:
        optional_args.add_argument("-s", "--setup_type", type=str, default='api', choices=['create', 'api'],
                                   help="Choose 'api' do load data from API (runtime: ~1min) [highly recommended], "
                                        "or 'create' to create it from scratch (runtime: ~3h) [Default=api]")

    optional_args.add_argument("-h", "--help", action="help", help="show this help message and exit")
    args = parser.parse_args()
    # ============================================================================
    # prepare input
    # ============================================================================
    if 'm' in arguments:
        if args.mode in ["set-set", "network-set"]:
            args.reference = pd.read_csv(args.reference, header=None, sep="\t", dtype=str)[0]
            args.reference = set(args.reference)
        if args.mode == "clustering":
            args.target = pd.read_csv(args.target, header=None, sep="\t", dtype=str, names=["id", "cluster", "desc"])
        else:
            args.target = pd.read_csv(args.target, header=None, sep="\t", dtype=str)[0]
            args.target = set(args.target)

    return args


def _get_epilog(script_name):
    epilog = ""
    if script_name == 'single_validation.py':
        epilog += "\n----------------------------------------------------------------------------\n"
        epilog += "\nsupported id types\n"
        epilog += "  for genes\t\t" + ', '.join(c.SUPPORTED_GENE_IDS) + "\n"
        epilog += "  for diseases\t\t" + ', '.join(c.SUPPORTED_DISEASE_IDS) + "\n"
        epilog += "\nsupported modes\n"
        epilog += "  set\t\t\tCompare similarity inside the set. Either genes or diseases.\n"
        epilog += "  set-set\t\tCompare target set to reference set. Both either genes or diseases.\n"
        epilog += "  clustering\t\tCompare cluster quality inside clustering. Either genes or diseases.\n"
        epilog += "  subnetwork\t\t\tCompare similarity inside the subnetwork nodes. Either genes or diseases.\n"
        epilog += "  subnetwork-set\t\tCompare target subnetwork to reference set. Both either genes or diseases.\n"
        epilog += "\nsupported background models\n"
        epilog += "  complete\t\tRandom ids will be picked fully randomized.\n"
        epilog += "  term-pres\t\tRandom ids will preserve the number of mapped terms for the replaced ids.\n"
        epilog += "  network\t\tRandom ids will preserve the number of connected components in given network.\n"
    epilog += "\n############################################################################\n"
    return epilog


def print_current_usage(text):
    memory_usage = '{0:.2f}'.format(psutil.Process(os.getpid()).memory_info().rss / 1024 ** 2)
    time_usage = time.strftime("%H:%M:%S", time.gmtime(time.time() - start_time))
    print('[{}|{}MB] '.format(time_usage, memory_usage) + text)
